import { getDb } from "#database";

const db = getDb("cache.sqlite");

db.table(/* SQL */ `
  CREATE TABLE IF NOT EXISTS permissions (
    prefix TEXT PRIMARY KEY,
    allow INTEGER NOT NULL
  );
`);

export class FilePermissions {
  prefix!: string;
  /** Currently set to 1 always */
  allow!: number;

  // -- static ops --

  static getByPrefix(filePath: string): number {
    return (getByPrefixQuery.get(filePath))?.allow ?? 0;
  }
  static getExact(filePath: string): number {
    return (getExactQuery.get(filePath))?.allow ?? 0;
  }
  static setPermissions(dirPath: string, level: number) {
    if (level) {
      insertQuery.run(dirPath, level);
    } else {
      deleteQuery.run(dirPath);
    }
  }
}

const getByPrefixQuery = db.prepare(/* SQL */ `
  SELECT allow 
  FROM permissions 
  WHERE ? GLOB prefix || '*'
  ORDER BY LENGTH(prefix) DESC 
  LIMIT 1;
`).type<FilePermissions>();

const getExactQuery = db.prepare(/* SQL */ `
  SELECT allow FROM permissions WHERE ? == prefix
`).type<FilePermissions>();

const insertQuery = db.prepare(/* SQL */ `
  REPLACE INTO permissions(prefix, allow) VALUES(?, ?);
`);
const deleteQuery = db.prepare(/* SQL */ `
  DELETE FROM permissions WHERE prefix = ?;
`);
