// __esModule
true

// importDefault
function importDefault(file, request, nameHint) {
  const imports = getImports(file);
  request = resolveRelativePath(file, request);
  let importDeclaration = imports.get(request);

  if (!importDeclaration) {
    imports.set(
      request,
      importDeclaration = file.path.pushContainer(
        "body",
        _compiler.types.importDeclaration([], _compiler.types.stringLiteral(request))
      )[0]
    );
  }

  if (!nameHint) {
    return;
  }

  const specifiers = importDeclaration.get("specifiers");
  const specifier = specifiers.find((specifier) =>
  specifier.isImportDefaultSpecifier()
  );

  if (!specifier) {
    const identifier = file.scope.generateUidIdentifier(nameHint);
    importDeclaration.pushContainer(
      "specifiers",
      _compiler.types.importDefaultSpecifier(identifier)
    );
    return identifier;
  }

  return _compiler.types.identifier(specifier.node.local.name);
}

// importNamed
function importNamed(file, request, name, nameHint = name) {
  request = resolveRelativePath(file, request);
  const imports = getImports(file);
  let importDeclaration = imports.get(request);

  if (!importDeclaration) {
    imports.set(
      request,
      importDeclaration = file.path.pushContainer(
        "body",
        _compiler.types.importDeclaration([], _compiler.types.stringLiteral(request))
      )[0]
    );
  }

  const specifiers = importDeclaration.get("specifiers");
  const specifier = specifiers.find(
    (specifier) =>
    specifier.isImportSpecifier() && specifier.node.imported.name === name
  );

  if (!specifier) {
    const identifier = file.scope.generateUidIdentifier(nameHint);
    importDeclaration.pushContainer(
      "specifiers",
      _compiler.types.importSpecifier(identifier, _compiler.types.identifier(name))
    );
    return identifier;
  }

  return _compiler.types.identifier(specifier.node.local.name);
}

// importStar
function importStar(file, request, nameHint) {
  const imports = getImports(file);
  request = resolveRelativePath(file, request);
  let importDeclaration = imports.get(request);

  if (!importDeclaration) {
    imports.set(
      request,
      importDeclaration = file.path.pushContainer(
        "body",
        _compiler.types.importDeclaration([], _compiler.types.stringLiteral(request))
      )[0]
    );
  }

  if (!nameHint) {
    return;
  }

  const specifiers = importDeclaration.get("specifiers");
  const specifier = specifiers.find((specifier) =>
  specifier.isImportNamespaceSpecifier()
  );

  if (!specifier) {
    const identifier = file.scope.generateUidIdentifier(nameHint);
    importDeclaration.pushContainer(
      "specifiers",
      _compiler.types.importNamespaceSpecifier(identifier)
    );
    return identifier;
  }

  return _compiler.types.identifier(specifier.node.local.name);
}

// resolveRelativePath
function resolveRelativePath(file, request) {
  if (request.startsWith(FS_START)) {
    request = (0, _relativeImportPath.relativeImportPath)(file.opts.filename, request);
  }

  if (file.markoOpts.optimize) {
    request = request.replace(
      /(^|\/node-modules\/)marko\/src\//,
      "$1marko/dist/"
    );
  }

  return request;
}
